##############################################################################
# Chapter 9                                                                  #
# Sanchez, J. Introduction to Time Series for data scientists, CUP 2023.                        #
# Code used to start exercise for  Chapter 9 case study 2 (not in book)      #
# Regression with time series features and arima residuals                    #
#  Uses some TSstudio,  and lubridate package                               #
#  Example inspired by Krispin [99]
# Program  Chapter9-UKgrid.R                                                 #
# Last modified: 2/26/2023                                                   #
##############################################################################


#install.packages("TSstudio")
 library(TSstudio)
#install.packages("UKgrid")
  library(UKgrid)
#install.packages("lubridate")
  library(lubridate)
#install.packages("dplyr")
   library(dplyr)
#install.packages("forecast")
library(forecast)


  class(UKgrid)  # a data frame 
  head(UKgrid)

### We aggregate to daily electricity demand with a function in UKgrid package, 
  ## the function extract_grid

  df = extract_grid(type="data.frame", 
                        columns="ND", 
                         aggregate="daily")
  head(df)
  class(df)
  names(df)=c("date", "y")
   str(df)
   head(df$y)
 
   
  
  # install.packaged("dygraphs")
  library(dygraphs)
  dygraph(df, main="UK demand for electricity", 
          ylab="Daily Demand for electicity (MW)") %>% dyRangeSelector
  # it looks like before 2010 the trend was flat, then 
  # after 2010 the trend's slope increased. 

  UKgrid
  
  
  ## We use the ts_heatmap function from TSstudio package
  ts_heatmap(df[which(year(df$date)>=2016),], 
       title="UK Daily electricity demand")
  
  
  class(df) # is a data frame 
  
  ### Features engineering
  
  #### Because the series has probably strong correlation with the 
  ## seasonal lags, and the time series is daily, Krispin used 
  # a lag of 365 observations. We use dplyr for that 
  
  df <- df %>% mutate(lag365=dplyr::lag(y,365)) %>%
    filter(!is.na(lag365)) %>% arrange(date)
  
  df$month= month(df$date, label=T)
  df$wday=wday(df$date, label= T)
  df$trend=  1:nrow(df) 
  head(df)
  acf(df$y)
  
  # notice that we lost 365 days because of the lag feature. The data now 
  # starts on 2007, April 1. 
  
  str(df)
  
 
  ## Split the data frame into a training set and a test set 
      
       train_df=df[1:(nrow(df)-31), ]
       test_df=df[(nrow(df)-31+1):nrow(df), ]
  
  ## Perform regression with features
       
    model1=lm(y~trend+month+wday+lag365, data=train_df)
     summary(model1)
     forecast=predict(model1, newdata=test_df)
   
      sdresiduals=(model1$residuals-mean(model1$residuals))/sd(model1$residuals)
   plot(sdresiduals, type="l")  
   acf(model1$residuals,main="Residuals are stationary but autocorrelated") 

   
##  Residuals are autocorrelated, so we need to model them. 
   
####### Start with forecast package procedures.We will use this package 
   ### instead of gls from nlme
   
  ## First, create a time series ts object of the output variable y
   
   ## April 1st of 2007 was the 91th day of the week and a Saturday. 
   
   y.ts=ts(df$y, frequency=365, start=c(2007,91))
  
  ## split into training and test sets. 
   
   par=ts_split(y.ts, sample.out=31) #let's just forecast the last month. 
   par  # not too big a dataset so we can view. 
   
   train=par$train 
   train
   test=par$test
   test
   head(train)
   
  ### Fit with auto.arima from forecast package to account
   #for the autocorrelated errors. Recall: auto.arima is an alternative to gls 
   
   model2=auto.arima(train, 
                     xreg=cbind(model.matrix(~cbind(month,wday),train_df )[,-1], 
                                train_df$lag365, 
                                train_df$trend), 
                                seasonal=TRUE, 
                                stepwise=FALSE, 
                                approximation=FALSE) 
   
   summary(model2)
   checkresiduals(model2)  # residuals have acf of white noise. 
   summary(model2)  
   
   
   #####################################################################
   # Question 2.                                                        # 
   # Compare the model that you chose in Question 1 and the results you #
   # obtained with it after you apply gls with the results obtained with #
   # the auto.arima function of the forecast package.                   #
   # Do you reach the same conclusions? Explain and compare             # 
   # ###################################################################
   
   
   
   fcast2 = forecast(model2, 
                xreg=cbind(model.matrix(~cbind(month,wday),test_df )[,-1],
                                        test_df$trend, 
                                        test_df$lag365))
   fcast2
   accuracy(fcast2, test)                             
   
   
   ### Obviously, looking at the residuals, the residuals of the model 
   ## resulting from auto.arima are not white noise. Create other features 
   ## that might help, for example the day, or perhaps a quadratic trend, 
   ## and modify the program slightly to add these other features. 
   ##Below are some suggestions on where you would have to start adding those. 
   ## after that, you make sure to try them in the models. 
   
       ### Suggested change of the df frame to add more features ####
      ## to add them to the models, change the model specification ### 
   
   df <- df %>% mutate(lag7=dplyr::lag(y,7)) %>%
     mutate(lag365=dplyr::lag(y,365)) %>%
     filter(!is.na(lag365)) %>% arrange(date)
   
   
   df$month= month(df$date, label=T)
   df$day=day(df$date)
   df$wday=wday(df$date, label= T)
   df$trend=  1:nrow(df) 
   df$trend2=df$trend^2
   head(df)
   acf(df$y)
   
   
   
   
   